/**********************************************************************

randlib.cpp

Methods for RandLib class.  See randlib.h for more info.

**********************************************************************/

#include "randlib.h"

// constructor

RandLib::RandLib()
{
   range_set = 0;
   seed = 0;
   bbs_initialized = 0;
   moa_initialized = 0;
   mtw_initialized = 0;
   rnc_initialized = 0;
   rsh_initialized = 0;
   rrw_initialized = 0;
   tpr_initialized = 0;
}

// algorithm methods

unsigned char RandLib::BBS_GetByte()
{
   double d;

   if (! bbs_initialized)
      return (unsigned char) 0;
   d = bbs.Random();

   return (unsigned char) AsInteger(d,0,255);
}

int RandLib::BBS_GetInteger()
{
   double d;

   if (! range_set)
      return (int) 0;
   if (! bbs_initialized)
      return (int) 0;
   d = bbs.Random();

   return AsInteger(d,min,max);
}

double RandLib::BBS_GetDouble()
{
   if (! bbs_initialized)
      return (double) 0.;
   return bbs.Random();
}

unsigned char RandLib::MOA_GetByte()
{
   uint32 s;
   double d;

   if (! moa_initialized)
   {
      s = GetSeed();
      moa.Init(s);
      moa_initialized = 1;
   }
   d = moa.Random();

   return (unsigned char) AsInteger(d,0,255);
}

int RandLib::MOA_GetInteger()
{
   uint32 s;
   double d;

   if (! range_set)
      return (int) 0;
   if (! moa_initialized)
   {
      s = GetSeed();
      moa.Init(s);
      moa_initialized = 1;
   }
   d = moa.Random();

   return AsInteger(d,min,max);
}

double RandLib::MOA_GetDouble()
{
   uint32 s;

   if (! moa_initialized)
   {
      s = GetSeed();
      moa.Init(s);
      moa_initialized = 1;
   }
   return moa.Random();
}

unsigned char RandLib::MTW_GetByte()
{
   uint32 s;
   double d;

   if (! mtw_initialized)
   {
      s = GetSeed();
      mtw.Init(s);
      mtw_initialized = 1;
   }
   d = mtw.Random();

   return (unsigned char) AsInteger(d,0,255);
}

int RandLib::MTW_GetInteger()
{
   uint32 s;
   double d;

   if (! range_set)
      return (int) 0;
   if (! mtw_initialized)
   {
      s = GetSeed();
      mtw.Init(s);
      mtw_initialized = 1;
   }
   d = mtw.Random();

   return AsInteger(d,min,max);
}

double RandLib::MTW_GetDouble()
{
   uint32 s;

   if (! mtw_initialized)
   {
      s = GetSeed();
      mtw.Init(s);
      mtw_initialized = 1;
   }
   return mtw.Random();
}

unsigned char RandLib::RNC_GetByte()
{
   uint32 s;
   double d;

   if (! rnc_initialized)
   {
      s = GetSeed();
      rnc.Init(s);
      rnc_initialized = 1;
   }
   d = rnc.Random();

   return (unsigned char) AsInteger(d,0,255);
}

int RandLib::RNC_GetInteger()
{
   uint32 s;
   double d;

   if (! range_set)
      return (int) 0;
   if (! rnc_initialized)
   {
      s = GetSeed();
      rnc.Init(s);
      rnc_initialized = 1;
   }
   d = rnc.Random();

   return AsInteger(d,min,max);
}

double RandLib::RNC_GetDouble()
{
   uint32 s;

   if (! rnc_initialized)
   {
      s = GetSeed();
      rnc.Init(s);
      rnc_initialized = 1;
   }
   return rnc.Random();
}

unsigned char RandLib::RSH_GetByte()
{
   uint32 s;
   double d;

   if (! rsh_initialized)
   {
      s = GetSeed();
      rsh.Init(s);
      rsh_initialized = 1;
   }
   d = rsh.Random();

   return (unsigned char) AsInteger(d,0,255);
}

int RandLib::RSH_GetInteger()
{
   uint32 s;
   double d;

   if (! range_set)
      return (int) 0;
   if (! rsh_initialized)
   {
      s = GetSeed();
      rsh.Init(s);
      rsh_initialized = 1;
   }
   d = rsh.Random();

   return AsInteger(d,min,max);
}

double RandLib::RSH_GetDouble()
{
   uint32 s;

   if (! rsh_initialized)
   {
      s = GetSeed();
      rsh.Init(s);
      rsh_initialized = 1;
   }
   return rsh.Random();
}

unsigned char RandLib::RRW_GetByte()
{
   uint32 s;
   double d;

   if (! rrw_initialized)
   {
      s = GetSeed();
      rrw.Init(s);
      rrw_initialized = 1;
   }
   d = rrw.Random();

   return (unsigned char) AsInteger(d,0,255);
}

int RandLib::RRW_GetInteger()
{
   uint32 s;
   double d;

   if (! range_set)
      return (int) 0;
   if (! rrw_initialized)
   {
      s = GetSeed();
      rrw.Init(s);
      rrw_initialized = 1;
   }
   d = rrw.Random();

   return AsInteger(d,min,max);
}

double RandLib::RRW_GetDouble()
{
   uint32 s;

   if (! rrw_initialized)
   {
      s = GetSeed();
      rrw.Init(s);
      rrw_initialized = 1;
   }
   return rrw.Random();
}

unsigned char RandLib::TPR_GetByte()
{
   uint32 s;
   double d;

   if (! tpr_initialized)
   {
      s = GetSeed();
      tpr.Init(s);
      tpr_initialized = 1;
   }
   d = tpr.Random();

   return (unsigned char) AsInteger(d,0,255);
}

int RandLib::TPR_GetInteger()
{
   uint32 s;
   double d;

   if (! range_set)
      return (int) 0;
   if (! tpr_initialized)
   {
      s = GetSeed();
      tpr.Init(s);
      tpr_initialized = 1;
   }
   d = tpr.Random();

   return AsInteger(d,min,max);
}

double RandLib::TPR_GetDouble()
{
   uint32 s;

   if (! tpr_initialized)
   {
      s = GetSeed();
      tpr.Init(s);
      tpr_initialized = 1;
   }
   return tpr.Random();
}

// initialization methods

void RandLib::SetSeed(uint32 the_seed)
{
   seed = the_seed;
}

void RandLib::SetIntegerRange(int the_min, int the_max)
{
   min = the_min;
   max = the_max;
   range_set = 1;
}

void RandLib::BBS_Init(verylong p, verylong q, verylong key)
{
   bbs.Init(p,q,key);
   bbs_initialized = 1;
}

void RandLib::BBS_BlumInit(verylong the_seed)
{
   if (ziszero(the_seed))
      zintoz(time(0),&the_seed);
   zrstart(the_seed);
}

verylong RandLib::BBS_GenBlumPrime(int numbits)
{
   verylong p = 0;
   long nnbits;

   nnbits = 0 - numbits;
   if (! zrandomprime(nnbits, 50, &p, zrandomb))
   {
      zzero(&p);
   }
   return p;
}

verylong RandLib::BBS_GenKey(verylong p, verylong q)
{
   verylong start = 0;
   verylong key = 0;
   verylong n = 0;
   verylong x = 0;
   verylong v = 0;
   long result, numbits;

   // find start value

   zmul(p,q,&n);
   numbits = z2log(n);
   start = BBS_GenBlumPrime(numbits);
   if (ziszero(start))
   {
      return start;
   }

   // now find key value

   zmod(start,n,&key);
   result = zscompare(key,(long) 1);
   if (result < 1)
      zintoz((long) 2,&key);

   zsq(key,&x);
   zmod(x,n,&key);
   zsq(x,&key);
   zmod(x,n,&key);
   zone(&v);

   zmod(key,p,&x);
   result = ziszero(x);
   if (result)
   {
      zzero(&v);
      zsadd(key, (long) 1, &x);
      zmod(x,n,&key);
   }
   zmod(key,q,&x);
   result = ziszero(x);
   if (result)
   {
      zzero(&v);
      zsadd(key, (long) 1, &x);
      zmod(x,n,&key);
   }
   result = zscompare(key, (long) 1);
   if (result < 1)
      zzero(&v);

   while (ziszero(v))
   {
      zsq(key,&x);
      zmod(x,n,&key);
      zsq(key,&x);
      zmod(x,n,&key);
      zone(&v);

      zmod(key,p,&x);
      result = ziszero(x);
      if (result)
      {
         zzero(&v);
         zsadd(key, (long) 1, &x);
         zmod(x,n,&key);
      }
      zmod(key,q,&x);
      result = ziszero(x);
      if (result)
      {
         zzero(&v);
         zsadd(key, (long) 1, &x);
         zmod(x,n,&key);
      }
      result = zscompare(key,(long) 1);
      if (result < 1)
         zintoz((long) 2,&key);
   }

   zfree(&start);
   zfree(&n);
   zfree(&x);
   zfree(&v);

   return key;
}

// private methods

uint32 RandLib::GetSeed()
{
   if (seed == 0)
      return time(0);
   return seed;
}

int RandLib::AsInteger(double d, int the_min, int the_max)
{
   int i, interval;

   interval = the_max - the_min + 1;
   i = (int) (d * interval);
   if (i > interval)
      i = interval;
   return (the_min + i);
}

